package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.Random;


/**
 * implements of Mix2ZeroKnowledgeProver interface
 * this implementation assumes that each RerandomizableEncryptedMessage is ElGamalCiphertext
 */
public class Prover implements Mix2ZeroKnowledgeProver {
//
    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final Random rand;
    private final ECElGamalEncryption encryptor;
    private final ECPoint g,h;
    private final BigInteger groupOrderUpperBound;
    private final ECElGamalMixStatementGenerator mixParams;
    final SigmaFiatShamir<Mixing.Mix2Proof, Mixing.Mix2Proof.FirstMessage, Mixing.Mix2Proof.FinalMessage> mix2NIZK;

    /**
     * @param rand
     * @param encryptor
     * @param randomOracle  - use for Fiat–Shamir heuristic
     */
    public Prover(Random rand, ECElGamalEncryption encryptor, RandomOracle randomOracle) {

        this.rand = rand;
        this.encryptor = encryptor;
        this.randomOracle = randomOracle;
        this.group = this.encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = this.encryptor.getElGamalPK().getPK();
        this.mixParams = new ECElGamalMixStatementGenerator(encryptor);
        this.groupOrderUpperBound  = group.orderUpperBound();
        this.mix2NIZK =  new SigmaFiatShamir(ProtobufConcatenators.concatNIZK, randomOracle);
    }

    /**
     * @param a
     * @param b
     * @param c  - if switched then c = rerandomize(b,r2) else c = rerandomize(a,r1)
     * @param d  - if switched then d = rerandomize(a,r1) else d = rerandomize(b,r2)
     * @param switched    - trueClauseIndex
     * @param layer - row of a,b in encryption table
     * @param r1
     * @param r2
     * @return      - a valid ZKP that indeed c,d calculated as required
     * @throws InvalidProtocolBufferException
     */
    public Mixing.Mix2Proof prove(Crypto.RerandomizableEncryptedMessage a,
                                           Crypto.RerandomizableEncryptedMessage b,
                                           Crypto.RerandomizableEncryptedMessage c,
                                           Crypto.RerandomizableEncryptedMessage d,
                                           boolean switched, int layer, int switchIdx, int out0Idx, int out1Idx,

                                           Crypto.EncryptionRandomness r1,
                                           Crypto.EncryptionRandomness r2) throws InvalidProtocolBufferException {


        ECElGamalMixStatementGenerator.Mix2Statement statement = mixParams.createStatement(a, b, c, d);
        ECElGamalMixStatementGenerator.Mix2StatementWitness witness = mixParams.createMix2Witness(r1, r2, switched);

        Mix2.Prover prover = new Mix2.Prover(encryptor, rand, statement, witness);

        Mixing.Mix2Proof.Location location = Mixing.Mix2Proof.Location.newBuilder()
                .setLayer(layer)
                .setSwitchIdx(switchIdx)
                .setOut0(out0Idx)
                .setOut1(out1Idx)
                .build();
        return mix2NIZK.generateNizk(prover).toBuilder().setLocation(location).build();
    }

}

