package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

/**
 * implements Mix2ZeroKnowledgeVerifier
 */
public class Verifier implements Mix2ZeroKnowledgeVerifier {


    private final ECElGamalEncryption encryptor;
    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final ECPoint g,h;
    private final ECElGamalMixStatementGenerator mixParams;
    final SigmaFiatShamir mix2NIZK;


//    private final ZeroKnowledgeOrProofParser parser;

    /**
     * constructor
     * @param encryptor should be as the encryptor used by meerkat.mixer.proofs
     * @param randomOracle should be as the random oracle used by meerkat.mixer.proofs
     */
    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.encryptor = encryptor;
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
        this.mixParams = new ECElGamalMixStatementGenerator(encryptor);
        this.mix2NIZK =  new SigmaFiatShamir(ProtobufConcatenators.concatNIZK, randomOracle);
//        this.parser = new ZeroKnowledgeOrProofParser(group);
    }

    /**
     * verify zero knowledge proof
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @param proof out1 = rerandomize(in1) && out2 = rerandomize(in2)
     *              ||
     *              out1 = rerandomize(in2) && out2 = rerandomize(in1)
     * @return  true iff all 4 or proofs are valid
     * @throws InvalidProtocolBufferException
     */
    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.Mix2Proof proof) throws InvalidProtocolBufferException {

        ECElGamalMixStatementGenerator.Mix2Statement statement = mixParams.createStatement(in1,in2,out1,out2);
        Mix2.Verifier verifier = new Mix2.Verifier(encryptor, statement);

        return mix2NIZK.verifyNizk(proof, verifier);
    }
}
